# -*- coding: utf-8 -*-
# Copyright (c) 2021, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.


class AndeboxException(Exception):
    pass


# code: language=python tabSize=4
